( function () {

	mw.urlshortener = {
		/**
		 * @member {mw.Api)
		 */
		api: new mw.Api(),

		/**
		 * @member {RegExp}
		 */
		regex: new RegExp( mw.config.get( 'wgUrlShortenerAllowedDomains' ) ),

		/**
		 * @member {boolean}
		 */
		allowArbitraryPorts: mw.config.get( 'wgUrlShortenerAllowArbitraryPorts' ),

		/**
		 * @member {OO.ui.TextInputWidget}
		 */
		shortened: null,

		/**
		 * @member {OO.ui.TextInputWidget}
		 */
		input: null,
		/**
		 * @member {OO.ui.ButtonInputWidget}
		 */
		submit: null,

		/**
		 * Validate the input URL clientside. Note that this is not the
		 * only check - they are checked serverside too.
		 *
		 * Checks for both URL validity and AllowedDomains matching.
		 *
		 * @param {string} input the URL that is to be shortened
		 * @return {boolean|Object} true if object is validated, an object matching what is
		 *                         returned by the API in case of error.
		 */
		validateInput: function ( input ) {
			var parsed,
				self = mw.urlshortener;

			try {
				parsed = new mw.Uri( input );
			} catch ( e ) {
				self.fieldLayout.setErrors( [ mw.msg( 'urlshortener-error-malformed-url' ) ] );
				return false;
			}
			if ( !parsed.host.match( self.regex ) ) {
				self.fieldLayout.setErrors( [ mw.msg( 'urlshortener-error-disallowed-url', parsed.host ) ] );
				return false;
			}
			if ( parsed.port &&
				!self.allowArbitraryPorts &&
				!( parsed.port === '80' || parsed.port === '443' )
			) {
				self.fieldLayout.setErrors( [ mw.msg( 'urlshortener-error-badports' ) ] );
				return false;
			}

			if ( parsed.user || parsed.password ) {
				self.fieldLayout.setErrors( [ mw.msg( 'urlshortener-error-nouserpass' ) ] );
				return false;
			}

			self.fieldLayout.setErrors( [] );
			return true;
		},

		/**
		 * Click handler for the submit button
		 */
		onSubmit: function () {
			var self = mw.urlshortener;
			self.input.getValidity().done( function () {
				self.input.pushPending().setReadOnly( true );
				self.setSubmit( 'submitting' );
				self.shortenUrl(
					self.input.getValue()
				).done( function ( shorturl ) {
					self.setSubmit( 'submit' );
					self.input.popPending().setReadOnly( false );

					if ( !self.shortened ) {
						self.shortened = new mw.widgets.CopyTextLayout( {
							align: 'top',
							label: mw.msg( 'urlshortener-shortened-url-label' ),
							copyText: shorturl,
							successMessage: mw.msg( 'urlshortener-copy-success' ),
							failMessage: mw.msg( 'urlshortener-copy-fail' )
						} );
						// Wrap in a FieldLayout so we get the label
						self.fieldLayout.$element.after( self.shortened.$element );
					} else {
						self.shortened.textInput.setValue( shorturl );
					}
					self.shortened.textInput.select();
				} ).fail( function ( err ) {
					self.setSubmit( 'submit' );
					self.input.popPending().setReadOnly( false );
					self.fieldLayout.setErrors( [ err.info ] );
				} );
			} );
		},

		init: function () {
			// eslint-disable-next-line no-jquery/no-global-selector
			var $field = $( 'form > .mw-htmlform-field-HTMLTextFieldWithButton' );
			if ( $field.length ) {
				this.fieldLayout = OO.ui.infuse( $field );
				this.input = this.fieldLayout.fieldWidget;
				this.input.setValidation( this.validateInput );
				this.submit = this.fieldLayout.buttonWidget;
				this.submit.on( 'click', this.onSubmit );
			}
		},

		/**
		 * @param {string} status either 'submitting' or 'submit'
		 */
		setSubmit: function ( status ) {
			// The following messages are used here:
			// * urlshortener-url-input-submitting
			// * urlshortener-url-input-submit
			this.submit.setLabel( mw.message( 'urlshortener-url-input-' + status ).text() );
		},

		/**
		 * Shorten the provided url
		 *
		 * @param {string} url
		 * @return {jQuery.Promise}
		 */
		shortenUrl: function ( url ) {
			return this.api.post( {
				action: 'shortenurl',
				url: url
			} ).then( function ( data ) {
				return data.shortenurl.shorturl;
			}, function ( errCode, data ) {
				return $.Deferred().reject( data.error ).promise();
			} );
		}
	};

	$( function () {
		mw.urlshortener.init();
	} );
}() );
